use std::{fmt::Display, path::PathBuf, sync::Arc};

use axum::{extract::{Path, State}, http::{StatusCode, Uri}, response::{Html, Redirect}, routing::{get, post}, Form, Router};
use ldap3::{LdapConn, LdapConnAsync, LdapConnSettings, Scope, SearchEntry};
use pandoc::{InputFormat, InputKind, MarkdownExtension, OutputKind};
use serde::{Deserialize, Serialize};
use sqlx::SqlitePool;
use strum::{EnumIter, IntoEnumIterator};
use tera::{Context, Tera};

use tokio::sync::Mutex;
use tower_http::services::ServeDir; 

use tracing::{debug, error, Level};
use uuid::Uuid;
use validator::Validate;

#[derive(Clone)]
struct AppState {
    tera: Tera,
    pool: SqlitePool,
    // conserve les données saisies pour les représenter au cas où elles ne sont pas validées
    form: Arc<Mutex<FormCandidate>>
}

impl AppState {
    async fn new() -> Self {
        let pool = sqlx::sqlite::SqlitePool::connect(DB_URL).await.unwrap();

        let tera = match Tera::new("./templates/**/*.html") {
            Ok(t) => {
                for tpl in t.get_template_names() {
                    debug!("Added template: {tpl}");
                }
                debug!("Template loaded");
                t
            }
            Err(e) => {
                error!("Parsing errors: {e}");
                println!("Parsing errors: {e}");
                ::std::process::exit(1);
            }
        };

        AppState {
            pool,
            tera,
            form: Arc::new(Mutex::new(FormCandidate::new()))
        }
    }
}

const DB_URL: &str = "sqlite://sqlite.db";

#[tokio::main]
async fn main() {
    tracing_subscriber::fmt().with_max_level(Level::DEBUG).init();

    let state = AppState::new().await;

    // build our application with a single route
    let app = Router::new()
        .route("/", get(inscription))
        .route("/login", get(login))
        .route("/login", post(logged_in))
        .route("/inscription", get(inscription))
        .route("/add_candidate", post(add_candidate))
        .route("/recapitulation/:uuid", get(recapitulation))
        .fallback(fallback)
        // Permet de servir les fichiers PDF des dossiers générés (static file)
        .nest_service("/dossiers", ServeDir::new("./dossiers"))
        // .nest("dossiers", get(get_static_file))
        .with_state(state);

    // run our app with hyper, listening globally on port 3000
    let listener = tokio::net::TcpListener::bind("0.0.0.0:3000").await.unwrap();
    axum::serve(listener, app).await.unwrap();
}

async fn fallback(uri: Uri) -> (StatusCode, String) {
    (StatusCode::NOT_FOUND, format!("No route for {uri}"))
}

async fn inscription(State(state): State<AppState>) -> Html<String> {
    debug!("Inscription");
    debug!("FormCandidate: {:?}", state.form);
    let form = state.form.lock().await;
    Html(state.tera.render(Template::Inscription.display(), &Context::from_serialize(form.clone()).unwrap()).unwrap())
}

async fn login(State(state): State<AppState>) -> Html<String> {
    debug!("Login");
    Html(state.tera.render(Template::Login.display(), &Context::new()).unwrap())
}

#[derive(Serialize, Deserialize, Validate, Debug, Clone)]
struct Login {
    login: String,
    password: String,
}

#[derive(Serialize, Deserialize, Validate, Debug, Clone, sqlx::FromRow)]
struct FormCandidate {
    #[validate(length(min = 5))]
    #[sqlx(rename="prenom")] 
    firstname: String,
    #[validate(length(min = 5))]
    #[sqlx(rename="nom")] 
    lastname: String,
    birthdate: String,
    birthplace: String,
    birthplace_dep: String,
    birthplace_country: String,
    citizenship: String,
    scholarship_holder: bool,
    bea_number: String, 
    address: String,
    postcode: String,
    city: String,
    country: String,
    mobile_phone: String,
    email: String,
    year_n: Option<AcademicYear>,
    year_n_minus_1: Option<AcademicYear>,
    year_n_minus_2: Option<AcademicYear>,
    year_n_minus_3: Option<AcademicYear>,
    #[serde(skip_deserializing)]
    errors: Vec<String>,
}

impl FormCandidate {
    fn new() -> FormCandidate {
        FormCandidate {
            firstname: String::new(),
            lastname: String::new(),
            birthdate: String::new(),
            birthplace: String::new(),
            birthplace_dep: String::new(),
            birthplace_country: String::new(),
            citizenship: String::new(),
            scholarship_holder: false,
            bea_number: String::new(), 
            errors: Vec::new(),
            address: String::new(),
            postcode: String::new(),
            city: String::new(),
            country: String::new(),
            mobile_phone: String::new(),
            email: String::new(),
            year_n: None,
            year_n_minus_1: None,
            year_n_minus_2: None,
            year_n_minus_3: None,
        }
    }
}

impl PartialEq for FormCandidate {
    fn eq(&self, other: &Self) -> bool {
        self.firstname == other.firstname && self.lastname == other.lastname
    }
}

#[derive(Clone,Serialize,Deserialize,Debug)]
enum YearOfBachelor {
    First,
    Second,
    Third
}

impl ToString for YearOfBachelor {
    fn to_string(&self) -> String {
        match self {
            YearOfBachelor::First => String::from("1ère année"),
            YearOfBachelor::Second => String::from("2ème année"),
            YearOfBachelor::Third => String::from("3ème année"),
        }
    }
}

#[derive(Clone,Serialize,Deserialize,Debug)]
struct AcademicYear {
    degree: String,
    year: YearOfBachelor,
    school: String
}

#[derive(Serialize, Deserialize, Validate, Debug, Clone, sqlx::FromRow)]
struct Candidate {
    #[sqlx(rename="prenom")] 
    firstname: String,
    #[sqlx(rename="nom")] 
    lastname: String,
    #[sqlx(rename="uuid")] 
    uuid: String,
}

#[derive(EnumIter)]
enum Template {
    Dossier,
    Inscription,
    Login,
    Pdf
}
impl Template {
    fn display(&self) -> &str {
        match self {
            Template::Pdf => "pdf.html",
            Template::Inscription => "inscription.html",
            Template::Login => "login.html",
            Template::Dossier => "dossier.html"
        }
    }
}

async fn logged_in(State(state): State<AppState>, Form(mut login): Form<Login>) -> Redirect {
    debug!("Logged in");
    let (conn, mut ldap) = LdapConnAsync::with_settings(LdapConnSettings::new()
        .set_starttls(false)
        .set_no_tls_verify(false),
        "ldaps://plg-dc2.ad.univ-lorraine.fr:636").await.unwrap();
    ldap3::drive!(conn);

    // Attempts a simple bind using the passed in values of username and Password
    let result = ldap.simple_bind("fery3@univ-lorraine.fr", "gutsberserk48").await.unwrap().success();
    if result.is_ok() {
        debug!("Auth OK");
    } else {
        debug!("Auth Failed");
    }
    // Ok(ldap.unbind());
    Redirect::to("/login")
}
async fn add_candidate(State(state): State<AppState>, Form(mut candidate): Form<FormCandidate>) -> Redirect {
    debug!("Add candidate");
    match candidate.validate() {
        Ok(_) => {
            let inserted_candidate = db_add_candidate(&state, candidate).await.unwrap();

            debug!("PDF candidate: {:?}", inserted_candidate);

            let rendered_data = state.tera.render(Template::Pdf.display(), &Context::from_serialize(&inserted_candidate).unwrap()).unwrap();

            let mut pandoc = pandoc::new();
            pandoc.set_input(InputKind::Pipe(rendered_data));
            pandoc.set_input_format(InputFormat::Html,[MarkdownExtension::RawHtml].to_vec());
            pandoc.set_output(OutputKind::File(PathBuf::from(format!("/home/fery3/opt/projectc/dossiers/{}.pdf", inserted_candidate.uuid))));
            pandoc.set_show_cmdline(true);
            let _ = pandoc.execute().unwrap();
            debug!("candidate validated");

            Redirect::to(&format!("/recapitulation/{}",inserted_candidate.uuid))
        },
        Err(e) => {
            for error in e.field_errors() {
                let (field, _) = error;
                candidate.errors.push(field.to_string());
            }
            debug!("add candidate error array {:?}", candidate.errors);

            let mut form = state.form.lock().await;
            *form = candidate.to_owned();
            // state.form = candidate;
            debug!("candidate not validated {e}");
            Redirect::to("/")
        }
    }
}

async fn recapitulation(State(state): State<AppState>, Path(uuid): Path<String>) -> Html<String> {

    let candidate = db_get_candidate(&state, uuid);
    Html(state.tera.render(Template::Dossier.display(), &Context::from_serialize(candidate.await.unwrap()).unwrap()).unwrap())
}

async fn db_add_candidate(state: &AppState, candidate: FormCandidate) -> Result<Candidate, sqlx::Error>{
    debug!("db_add_candidate: {} {}", candidate.firstname, candidate.lastname);
    let uuid = Uuid::new_v4();
    let insert = sqlx::query("insert into users (prenom, nom, uuid) values ($1,$2,$3)")
        .bind(&candidate.firstname)
        .bind(&candidate.lastname)
        .bind(uuid.to_string())
        .execute(&state.pool).await;
    match insert {
        Ok(r) => {
            let userid = r.last_insert_rowid();
            sqlx::query_as::<_, Candidate>("select prenom, nom, uuid from users where id = $1")
                .bind(userid)
                .fetch_one(&state.pool).await
        }
        Err(e) => {
            error!("add_candidate error: {}", e);
            Err(e)
        }
    }
}
async fn db_get_candidate(state: &AppState, token: String) -> Result<Candidate, sqlx::Error>{
    let candidate = sqlx::query_as::<_, Candidate>("select nom, prenom, uuid from users where uuid = $1")
        .bind(token)
        .fetch_one(&state.pool).await;
    match candidate {
        Ok(c) => Ok(c),
        Err(e) => {
            error!("add_candidate error: {}", e);
            Err(e)
        }
    }
}

// #[cfg(test)]
// #[tokio::test]
// async fn test_sql_candidate() {
//     let state = AppState::new().await;
//     let candidate = FormCandidate {
//         lastname: "Fery".to_string(),
//         firstname: "Yann".to_string(),
//     };
//     let inserted_candidate = db_add_candidate(&state, candidate.clone()).await.unwrap();
//     assert_eq!(candidate, inserted_candidate);

// }

// #[tokio::test]
// async fn test_template_list() {
//     let state = AppState::new().await;

//     let mut tera: Vec<&str> = [].to_vec();
//     for tpl in state.tera.get_template_names() {
//         tera.push(tpl);
//     }
//     let mut list: Vec<String> = [].to_vec();
//     for tpl in Template::iter() {
//         list.push(tpl.display().to_string());
//     }
//     assert_eq!(tera, list);
// }
